@@ -11,6 +11,11 @@
 namespace Quotient {
 
 struct RoomEventFilter : EventFilter {
+    /// If `true`, enables per-[thread](/client-server-api/#threading)
+    /// notification counts. Only applies to the `/sync` endpoint. Defaults to
+    /// `false`.
+    Omittable<bool> unreadThreadNotifications;
+
     /// If `true`, enables lazy-loading of membership events. See
     /// [Lazy-loading room
     /// members](/client-server-api/#lazy-loading-room-members) for more
@@ -44,6 +49,8 @@ struct JsonObjectConverter<RoomEventFilter> {
     static void dumpTo(QJsonObject& jo, const RoomEventFilter& pod)
     {
         fillJson<EventFilter>(jo, pod);
+        addParam<IfNotEmpty>(jo, QStringLiteral("unread_thread_notifications"),
+                             pod.unreadThreadNotifications);
         addParam<IfNotEmpty>(jo, QStringLiteral("lazy_load_members"),
                              pod.lazyLoadMembers);
         addParam<IfNotEmpty>(jo, QStringLiteral("include_redundant_members"),
@@ -56,6 +63,8 @@ struct JsonObjectConverter<RoomEventFilter> {
     static void fillFrom(const QJsonObject& jo, RoomEventFilter& pod)
     {
         fillFromJson<EventFilter>(jo, pod);
+        fromJson(jo.value("unread_thread_notifications"_ls),
+                 pod.unreadThreadNotifications);
         fromJson(jo.value("lazy_load_members"_ls), pod.lazyLoadMembers);
         fromJson(jo.value("include_redundant_members"_ls),
                  pod.includeRedundantMembers);
