@@ -4,12 +4,12 @@
 
 #pragma once
 
-#include "events/eventloader.h"
+#include "events/roomevent.h"
 #include "jobs/basejob.h"
 
 namespace Quotient {
 
-/*! \brief Listen on the event stream.
+/*! \brief Listen on the event stream of a particular room.
  *
  * This will listen for new events related to a particular room and return
  * them to the caller. This will block until an event is received, or until
@@ -22,9 +22,9 @@ namespace Quotient {
  * API will also be deprecated at some point, but its replacement is not
  * yet known.
  */
-class PeekEventsJob : public BaseJob {
+class QUOTIENT_API PeekEventsJob : public BaseJob {
 public:
-    /*! \brief Listen on the event stream.
+    /*! \brief Listen on the event stream of a particular room.
      *
      * \param from
      *   The token to stream from. This token is either from a previous
