@@ -13,7 +13,7 @@ namespace Quotient {
  * Sets the position of the read marker for a given room, and optionally
  * the read receipt's location.
  */
-class SetReadMarkerJob : public BaseJob {
+class QUOTIENT_API SetReadMarkerJob : public BaseJob {
 public:
     /*! \brief Set the position of the read marker for a room.
      *
@@ -26,11 +26,18 @@ public:
      *
      * \param mRead
      *   The event ID to set the read receipt location at. This is
-     *   equivalent to calling ``/receipt/m.read/$elsewhere:example.org``
+     *   equivalent to calling `/receipt/m.read/$elsewhere:example.org`
+     *   and is provided here to save that extra call.
+     *
+     * \param mReadPrivate
+     *   The event ID to set the *private* read receipt location at. This
+     *   equivalent to calling `/receipt/m.read.private/$elsewhere:example.org`
      *   and is provided here to save that extra call.
      */
-    explicit SetReadMarkerJob(const QString& roomId, const QString& mFullyRead,
-                              const QString& mRead = {});
+    explicit SetReadMarkerJob(const QString& roomId,
+                              const QString& mFullyRead = {},
+                              const QString& mRead = {},
+                              const QString& mReadPrivate = {});
 };
 
 } // namespace Quotient
